@@ -1,31 +1,21 @@
-import datetime
+import copy
 import logging
 import multiprocessing as mp
 import os
-import time
 from collections import defaultdict
+from collections.abc import Mapping
+
 from torch.utils.data import DataLoader
 
-from anc_data.anc_dataset import AncDataset
+from anc_data.anc_dataset import DATASET_IDX, AncDataset
 from anc_data.anc_processor import AncProcessor
 from anc_data.anc_sampler import AncMultiSourceSampler, AncSampler
-from anc_data.utils import DataNotReady
 
 
 def fake_collate(x):
     return x
 
 
-def safe_qsize(queue):
-    """Safely get queue size, handling NotImplementedError on macOS"""
-    try:
-        return queue.qsize()
-    except NotImplementedError:
-        # On macOS, qsize() is not implemented, so we return a default value
-        # that won't affect the logging logic significantly
-        return 0
-
-
 class AncDataLoader:
     r"""
     A wrapper for dataloader, dataset and sampler
@@ -91,7 +81,10 @@ class AncDataLoader:
         # TODO: add support to get rank and world from torch env
         self.paths = paths
         self.batch_size = batch_size
-        self.num_workers = num_workers
+        # set self.num_workers to 1 if num_workers=0
+        # torch loader will use num_workers and other part will use self.num_workers
+        self.num_workers = num_workers if num_workers > 0 else 1
+        self.repeat = repeat
         if num_workers > 0:
             if use_spawn:
                 ctx = mp.get_context("spawn")
@@ -100,6 +93,13 @@ class AncDataLoader:
                 self.ds_state_queues = [mp.Queue() for _ in range(num_workers)]
         else:
             self.ds_state_queues = [mp.Queue()]
+            prefetch_factor = None
+            if ckpt_interval is not None:
+                logging.warning(
+                    "ckpt is not supported for num_worker = 0, will set ckpt_interval to None"
+                )
+                ckpt_interval = None
+
         if int(os.getenv("ANC_DISABLE_LOGGING", "0")) == 1:
             enable_logging = False
         ds_ckpt_interval = ckpt_interval
@@ -135,7 +135,7 @@ class AncDataLoader:
                 batch_size=batch_size,
                 world=world,
                 rank=rank,
-                num_workers=num_workers,
+                num_workers=self.num_workers,
                 shuffle=shuffle,
                 seed=seed,
                 drop_last=drop_last,
@@ -150,7 +150,7 @@ class AncDataLoader:
                 batch_size=batch_size,
                 world=world,
                 rank=rank,
-                num_workers=num_workers,
+                num_workers=self.num_workers,
                 shuffle=shuffle,
                 seed=seed,
                 drop_last=drop_last,
@@ -174,8 +174,8 @@ class AncDataLoader:
         self.rank = rank
         self.world = world
         self.loader_iter = None
-        self.buffer = [[] for _ in range(num_workers)]
-        self.last_wid = -1
+        self.buffer = [[] for _ in range(self.num_workers)]
+        self.cur_wid = self.step % self.num_workers
         self.out_of_data_worker = set()
         self.max_steps = max_steps
         self.logging = enable_logging
@@ -187,6 +187,8 @@ class AncDataLoader:
         self.first_batch = None
         # with multi val dataloader, we may get _loader_idx as _ds_idx in data, so use a dict instead of a list here
         self.ds_read_counts = defaultdict(int)
+        self.no_more_data = False
+        self.all_ranks_no_more_data = False
 
     def __iter__(self):
         logging.debug(f"pid {os.getpid()} rank {self.rank} in iter")
@@ -200,12 +202,15 @@ class AncDataLoader:
             return self.max_steps
         return len(self.sampler)
 
-    def _get_next_wid(self, cur_wid):
+    def _try_advance_wid_to_next_unfinished(self):
+        if self.cur_wid is None:
+            return False
         for i in range(self.num_workers):
-            cur_wid = (cur_wid + 1) % self.num_workers
-            if cur_wid not in self.out_of_data_worker:
-                break
-        return cur_wid
+            if self.cur_wid not in self.out_of_data_worker or self.buffer[self.cur_wid]:
+                return True
+            self.cur_wid = (self.cur_wid + 1) % self.num_workers
+        self.cur_wid = None
+        return False
 
     def _empty_buffer(self):
         logging.info(
@@ -217,82 +222,73 @@ class AncDataLoader:
                 return data
         return None
 
+    def _try_getting_next_batch_for_curr_wid(self):
+        if self.cur_wid is None:
+            assert self.no_more_data, "cur_wid is None but no_more_data is False"
+            return None
+        for loop_cnt in range(self.num_workers * 1000):
+            if self.buffer[self.cur_wid]:
+                return self.buffer[self.cur_wid].pop()
+            if self.cur_wid in self.out_of_data_worker:
+                # no more data will come from cur_wid
+                return None
+
+            # pytorch dataloader with default batch_size=1 when not specified
+            cur_data_wid, cur_data, is_last_batch = next(self.loader_iter)[0]
+            if is_last_batch:
+                # curr_data_wid's sampler yielded no more indices
+                assert cur_data is None
+                self.out_of_data_worker.add(cur_data_wid)
+            else:
+                self.buffer[cur_data_wid].append(cur_data)
+
+        logging.warning(
+            f"{self.rank} Too many loop ({loop_cnt} times) when getting next data"
+            f"for worker {self.cur_wid} of step {self.step}"
+        )
+        return None
+
     def __next__(self):
+        if self.all_ranks_no_more_data:
+            raise StopIteration
         if self.always_return_first_batch and self.first_batch is not None:
             return self.first_batch
-        now = datetime.datetime.now()
-        need_logging = safe_qsize(self.loader_iter._data_queue) < self.num_workers
-        t1 = time.time()
-        data = None
-        if len(self.out_of_data_worker) == self.num_workers:
-            # TODO: better handle the exception
-            data = self._empty_buffer()
-            if data is None:
-                raise StopIteration
-        # decide which subprocess to get data from
-        if self.last_wid < 0:
-            cur_wid = self.step % self.num_workers
-        else:
-            cur_wid = (self.last_wid + 1) % self.num_workers
-        if cur_wid in self.out_of_data_worker:
-            cur_wid = self._get_next_wid(cur_wid)
-        if data is None and len(self.buffer[cur_wid]) > 0:
-            data = self.buffer[cur_wid].pop()
-        elif data is None:
-            cnt = 0
-            while True:
-                item = next(self.loader_iter)
-                # since we set batch size=1 for torch dataloader,
-                # the actual batch size is handled by sampler
-                # we need to use index 0 to get data out
-                # note: cur_data here should be a list of batch data
-                cur_data_wid, cur_data, is_last_batch = item[0]
-                if is_last_batch:
-                    self.out_of_data_worker.add(cur_data_wid)
-                    if len(self.out_of_data_worker) == self.num_workers:
-                        data = self._empty_buffer()
-                        if data is None:
-                            raise StopIteration
-                        else:
-                            break
-                if isinstance(cur_data, DataNotReady):
-                    logging.info(
-                        f"next call get DataNotReady from {cur_data_wid}, is last batch {is_last_batch}"
-                    )
-                    if is_last_batch:
-                        cur_wid = self._get_next_wid(cur_wid)
-                    continue
-                assert isinstance(cur_data, list)
-                if cur_data_wid == cur_wid:
-                    data = cur_data[0]
-                    self.buffer[cur_data_wid] += cur_data[1:]
-                    break
-                else:
-                    # store the out of order data
-                    self.buffer[cur_data_wid] += cur_data
-                cnt += 1
-                if cnt % self.num_workers == 0:
-                    logging.warning(
-                        f"{self.rank} Too many loop ({cnt} times) when getting next data"
-                        " for worker {cur_wid} of step {self.step}"
-                    )
-        self.step += 1
-        self.last_wid = cur_wid
-        consumed_time = time.time() - t1
-        need_logging = self.logging and (need_logging or consumed_time > 0.01)
-        if need_logging:
-            logging.debug(
-                f"__next__ called time {now} pid {os.getpid()} rank {self.rank} return step {self.step}"
-                " consumed time {consumed_time} data with buffer {safe_qsize(self.loader_iter._data_queue)}"
-            )
-        if self.always_return_first_batch and self.first_batch is None:
-            self.first_batch = data
 
-        if "_ds_idx" in data:
-            for idx in data["_ds_idx"]:
+        batch = None
+        for loop_cnt in range(self.num_workers * 1000):
+            batch = self._try_getting_next_batch_for_curr_wid()
+            if batch is not None:
+                # prepare for the next wid in round robin fashion
+                self.cur_wid = (self.cur_wid + 1) % self.num_workers
+                break
+            if not self._try_advance_wid_to_next_unfinished():
+                # all buffers are empty and all workers are out of data, mark permanently done for this rank
+                self.no_more_data = True
+                break
+            if loop_cnt > 0 and loop_cnt % 20 == 0:
+                logging.warning(
+                    f"rank {self.rank} is still getting data from worker {self.cur_wid} of step {self.step}"
+                )
+        if loop_cnt >= self.num_workers * 1000:
+            raise RuntimeError(
+                f"Too many loop ({loop_cnt} times) when getting next data for worker {self.cur_wid} of step {self.step}"
+            )
+
+        self.step += 1
+        if self.no_more_data:
+            logging.info(
+                f"rank {self.rank} has no more data, will keep returning the first batch"
+            )
+            return copy.deepcopy(self.first_batch)
+
+        assert batch is not None
+        if isinstance(batch, Mapping) and DATASET_IDX in batch:
+            for idx in batch[DATASET_IDX]:
                 self.ds_read_counts[idx] += 1
-            del data["_ds_idx"]
-        return data
+        if self.first_batch is None:
+            self.first_batch = copy.deepcopy(batch)
+
+        return batch
 
     def get_checkpoint(self):
         if self.step in self.last_ckpt:
@@ -327,10 +323,10 @@ class AncDataLoader:
         state["paths"] = self.paths
         state["batch_size"] = self.batch_size
         state["num_workers"] = self.num_workers
-        state["last_wid"] = self.last_wid
-        state["out_of_data_worker"] = self.out_of_data_worker
+        state["cur_wid"] = self.cur_wid
+        state["out_of_data_worker"] = copy.deepcopy(self.out_of_data_worker)
         state["step"] = self.step
-        state["ds_read_counts"] = self.ds_read_counts[:]  # Create a copy of the array
+        state["ds_read_counts"] = copy.deepcopy(self.ds_read_counts)
         return state
 
     def __setstate__(self, state):
