@@ -128,6 +128,8 @@ class RangesHolder:
         for i in other.list_of_ranges:
             self.list_of_ranges.append(i)
         self.length = sum(len(r) for r in self.list_of_ranges)
+        # Recompute accumulated lengths after modifying ranges
+        self.accu_lengths = self.create_accu_length(self.list_of_ranges)
 
     def to_list(self, shuffle=False):
         arrays = []
@@ -285,6 +287,7 @@ class AncSampler(DistributedSampler):
         self.shuffle = shuffle
         if getattr(self, "dataset", None) is not None:
             self.dataset = None
+        self.padding_indices = set()
 
     def _get_indices_with_ratio(self, indices, ratio):
         ratio_int = math.floor(ratio)
@@ -376,7 +379,7 @@ class AncSampler(DistributedSampler):
         indices = np.concatenate(indices_list)
         return indices
 
-    def _create_chunk_indices_from_bz_and_worker(self, indices):
+    def _compute_worker_indices_ranges(self):
         # calculate num batches per worker
         num_batches = (self.num_samples + self.batch_size - 1) // self.batch_size
         last_batch_size = self.num_samples - (num_batches - 1) * self.batch_size
@@ -403,6 +406,12 @@ class AncSampler(DistributedSampler):
             start_idx_per_worker[i] = (
                 start_idx_per_worker[i - 1] + num_samples_per_worker[i - 1]
             )
+        return num_batches, num_samples_per_worker, start_idx_per_worker
+
+    def _create_chunk_indices_from_bz_and_worker(self, indices):
+        num_batches, num_samples_per_worker, start_idx_per_worker = (
+            self._compute_worker_indices_ranges()
+        )
 
         if self.repeat and self.shuffle:
             self._np_random = np.random.default_rng(self.seed)
@@ -449,6 +458,9 @@ class AncSampler(DistributedSampler):
                         (indices[shift_offset:], indices[:shift_offset])
                     )
 
+    def is_padding(self, ds_idx, index):
+        return index in self.padding_indices
+
     def _iter_init(self):
         indices = []
         if self.epoch == self.last_iter_epoch:
@@ -474,9 +486,18 @@ class AncSampler(DistributedSampler):
                 else:
                     indices = RangesHolder(sub_indices_list)
             else:
-                indices = self._get_indices(self.sub_lengths, ratio=self.ratios)
-                if self.global_shuffle:
-                    self._np_random.shuffle(indices)
+                if not self.mem_save_mode:
+                    indices = self._get_indices(self.sub_lengths, ratio=self.ratios)
+                    if self.global_shuffle:
+                        self._np_random.shuffle(indices)
+                else:
+                    # Build a RangesHolder when mem_save_mode=True for flat sub_lengths
+                    sub_indices_list = self._get_indices(
+                        self.sub_lengths, return_sub_indices=True, ratio=self.ratios
+                    )
+                    if self.shuffle:
+                        self._np_random.shuffle(sub_indices_list)
+                    indices = RangesHolder(sub_indices_list)
 
             # if max_samples is set, may need to truncate indices to total_size
             if self.drop_last or len(indices) >= self.total_size:
@@ -485,6 +506,7 @@ class AncSampler(DistributedSampler):
                     indices = indices.slice(0, self.total_size)
                 else:
                     indices = indices[: self.total_size]
+                padding_size = 0  # No padding when truncating
             else:
                 # add extra samples to make it evenly divisible
                 padding_size = self.total_size - len(indices)
@@ -503,6 +525,20 @@ class AncSampler(DistributedSampler):
                     indices = np.concatenate((indices, padding))
             assert len(indices) == self.total_size
 
+            # max padding indices len: min(world_size, num_samples), should be small enough
+            if (
+                padding_size > 0
+                and len(indices) - padding_size < (self.rank + 1) * self.num_samples
+            ):
+                start = len(indices) - padding_size
+                end = (self.rank + 1) * self.num_samples
+                if self.mem_save_mode:
+                    self.padding_indices = set(
+                        indices.slice(start, end - start).to_list().tolist()
+                    )
+                else:
+                    self.padding_indices = set(indices[start:end])
+
             # we do the chunk sharding here
             if self.mem_save_mode:
                 indices = indices.slice(self.rank * self.num_samples, self.num_samples)
@@ -619,6 +655,9 @@ class AncMultiSourceSampler:
             indices, is_last_batch = next(cur_iter)
             yield cur_iter_idx, indices, is_last_batch
 
+    def is_padding(self, ds_idx, index):
+        return self.samplers[ds_idx].is_padding(ds_idx, index)
+
     def __len__(self):
         return 1000000000000
 
