@@ -636,7 +636,11 @@ class SFTProcessor(PretrainProcessor):
             dialog_sep: str = self.opt.delimiter
 
         def _encode_one(
-            prefix: str, content: str, suffix: str, is_output: bool = False
+            prefix: str,
+            content: str,
+            suffix: str,
+            is_output: bool = False,
+            is_last_round=False,
         ):
             """Helper function to encode a single turn of conversation.
 
@@ -649,6 +653,8 @@ class SFTProcessor(PretrainProcessor):
             Returns:
                 tuple: (role_tokens, content_tokens, suffix_tokens)
             """
+            if is_last_round:
+                suffix = suffix.rstrip()
             if is_output or no_query_mask:
                 # For assistant outputs or when no_query_mask is True,
                 # encode role prefix separately from content
@@ -673,6 +679,7 @@ class SFTProcessor(PretrainProcessor):
         # Process each turn in the conversation
         for i, (role, content) in enumerate(context):
             suffix_sep = dialog_sep
+            is_last_round = i == len(context) - 1
 
             if role == "assistant":
                 # Handle assistant's responses
@@ -681,11 +688,12 @@ class SFTProcessor(PretrainProcessor):
                     content,
                     suffix_sep,
                     is_output=True,
+                    is_last_round=is_last_round,
                 )
                 # Set mask based on loss_on_last_turn flag
-                content_mask = [
-                    int(not loss_on_last_turn or i == len(context) - 1)
-                ] * len(encoded_content)
+                content_mask = [int(not loss_on_last_turn or is_last_round)] * len(
+                    encoded_content
+                )
             else:
                 # Handle other roles (user, system, etc.)
                 if role not in (
@@ -701,7 +709,10 @@ class SFTProcessor(PretrainProcessor):
                 ):
                     logging.warn(f"Found unknown role={role}")
                 encoded_role, encoded_content, encoded_suffix = _encode_one(
-                    prefix_template.format(role=role), content, suffix_sep
+                    prefix_template.format(role=role),
+                    content,
+                    suffix_sep,
+                    is_last_round=is_last_round,
                 )
                 content_mask = [int(no_query_mask)] * len(encoded_content)
 
